import os, sys
import subprocess
import pytest
import pya

import lymask
from lymask import batch_main
from lytest import run_xor

from conftest import test_dir
dataprep_file = os.path.join(test_dir, 'tech', 'lymask_example_tech', 'dataprep', 'default.yml')
layout_file = os.path.join(test_dir, '1_src.oas')
outfile = os.path.join(test_dir, '1_run.oas')
reffile = os.path.join(test_dir, '1_answer.oas')


# This one need Technology working
def test_api():
    lymask.set_active_technology('lymask_example_tech')
    batch_main(layout_file, ymlspec=dataprep_file, outfile=outfile)
    run_xor(outfile, reffile)


def test_lyp_loading():
    from lymask.utilities import set_active_technology, reload_lys, lys
    layout = pya.Layout()
    layout.read(layout_file)
    lys.active_layout = layout

    lymask.set_active_technology('lymask_example_tech')
    lymask.utilities.reload_lys()

    assert lys['m5_wiring'] is lys('m5_wiring') is lys.m5_wiring

    with pytest.raises(KeyError):
        batch_main(layout_file, ymlspec='bad_masks', technology='lymask_example_tech')


def test_from_technology():
    batch_main(layout_file, ymlspec='default', outfile=outfile, technology='lymask_example_tech')
    run_xor(outfile, reffile)


def test_cm_from_tech():
    # this also checks that it defaults to default.yml
    command = ['lymask', 'dataprep']
    command += [layout_file]
    command += ['-o', outfile]
    command += ['-t', 'lymask_example_tech']
    subprocess.check_call(command)
    run_xor(outfile, reffile)




__version__ = '0.1.11'
__lygadget_link__ = ['lygadgets', 'yaml']

from lymask.utilities import active_technology, set_active_technology
from lymask.invocation import batch_main, batch_drc_main
from lymask.utilities import get_design_rules
import lymask.dataprep_steps



from __future__ import division, print_function, absolute_import
from functools import wraps
import os
import importlib.util

from lygadgets import pya, isGUI, message, message_loud
from lygadgets import anyCell_to_anyCell
from lygadgets.gui_objects import gui_view

from lymask.utilities import lys, LayerSet, active_technology, func_info_to_func_and_kwargs
from lymask.library import dbu, as_region, fast_sized, fast_smoothed, set_threads


all_dpfunc_dict = {}
def dpStep(step_fun):
    ''' Each step must accept one argument that is cell, plus optionals, and not return

        steps are added to all_steps *in the order they are defined*
    '''
    all_dpfunc_dict[step_fun.__name__] = step_fun
    return step_fun


def delete_dollar_duplicates(cell):
    ''' pya sometimes adds a second top cell or duplicates of other cells.
        Other layout programs and lithography tools cannot parse this easily.
        These all end with '$1', so they can be removed recursively
    '''
    layout = cell.layout()
    if cell.name.endswith('$1'):
        try:
            cell.delete()
        except RuntimeError as err:
            if 'mp_v->is_used (m_n)' in err.args[0]:
                pass
            elif 'Not a valid cell index' not in err.args[0]:
                print('BAD CELL INDEX', chop_cell.name)
                pass
            else:
                raise
        return
    to_delete = set()
    for child_id in cell.each_child_cell():
        child_cell = layout.cell(child_id)
        delete_dollar_duplicates(child_cell)


def dpStep_phidl(step_fun):
    ''' phidl version, where the mutable object is a phidl.Device not a pya.Cell
        Each step must accept one argument that is Device, plus optionals, and not return
    '''
    try:
        from phidl import Device
    except ImportError as err:
        raise ImportError('You are probably trying to use phidl dataprep steps within a GUI. This is only supported in batch mode. Not my fault')
    @wraps(step_fun)
    def wrapper(cell, *args, **kwargs):
        phidl_device = Device()
        anyCell_to_anyCell(cell, phidl_device)
        step_fun(phidl_device, *args, **kwargs)
        anyCell_to_anyCell(phidl_device, cell)
        delete_dollar_duplicates(cell)
    all_dpfunc_dict[step_fun.__name__] = wrapper
    return wrapper

# @dpStep_phidl
# def phidl_example(device):
#     import phidl.geometry as pg
#     device << pg.rectangle((10, 10))


@dpStep
def add_library(cell, filename):
    ''' Imports from the filename, which is a path to a python file.
        Anything within there that is a dpStep gets added to the all_dpfunc_dict for later
    '''
    if os.path.isfile(filename):
        pass
    else:
        dataprep_relpath = os.path.join(active_technology().eff_path('dataprep'), filename)
        if os.path.isfile(dataprep_relpath):
            filename = os.path.realpath(dataprep_relpath)
        else:
            raise FileNotFoundError('lymask could not find {}'.format(filename))
    modulename = os.path.splitext(os.path.basename(filename))[0]
    spec = importlib.util.spec_from_file_location(modulename, filename)
    foo = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(foo)
    except ImportError as err:
        if 'gdspy' in err.args[0]:
            raise ImportError('You are probably trying to use phidl/gdspy dataprep steps within a GUI. This is only supported in batch mode currently')
        else:
            raise


@dpStep
def check_floorplan(cell, fp_safe=50):
    ''' Checks for floorplan. If you didn't make one, this makes one.
    '''
    if cell.shapes(lys.FLOORPLAN).is_empty():
        message('Warning: No floorplan found. Making one, but you should do this yourself')
        fp_box = cell.dbbox()  # this assumes that, if FLOORPLAN is present, that DRC has verified it forms the extent
        fp_box.enlarge(fp_safe, fp_safe)
        cell.shapes(lys.FLOORPLAN).insert(fp_box)


__warned_about_flattening = False
@dpStep
def flatten(cell):
    global __warned_about_flattening
    if isGUI() and not __warned_about_flattening:
        message_loud('Warning: The flattening step modifies the layout, so be careful about saving.')
        __warned_about_flattening = True
    cell.flatten(True)


@dpStep
def paths_to_polys(cell):
    for layname in lys.keys():
        lay = lys[layname]
        for shape in cell.each_shape(lay):
            if shape.is_path():
                shape.polygon = shape.simple_polygon


@dpStep
def erase_text_and_other_junk(cell):
    for layname in lys.keys():
        lay = lys[layname]
        for text in cell.shapes(lay).each(pya.Shapes.STexts):
            cell.shapes(lay).erase(text)
        # zero width paths


# Some peole use this, others have already converted
# @dpStep
# def convert_wgs(cell):
    # has_Si = pya.Region()
    # for si in ['wg_deep', 'wg_shallow']:
    #     has_Si.insert(cell.shapes(lys[si]))
    # ^ new version-ish

#     union(cell, 'wg_deep', 'wg_shallow')
#     layer_bool(cell, lys.wg_deep, lys.wg_shallow,
#                pya.EdgeProcessor.ModeBNotA, layC=lys.wg_shallow)
#     cell.clear(lys.wg_deep)
#     cell.move(lys.dp_temp, lys.wg_deep)


# @dpStep
# def nanowire_heal(cell):
#     delta = 0.01 / dbu
#     layer_resize(cell, lys.m2_nw, delta, layC=lys.dp_temp)
#     layer_resize(cell, lys.dp_temp, -delta, layC=lys.m2_nw)


@dpStep
def processor(cell, thread_count=1, tiles=2, remote_host=None):
    if remote_host is not None:
        message_loud('Automatic remote hosting is not yet supported')
    set_threads(thread_count, tiles)


@dpStep
def nanowire_sleeve(cell, Delta=2.5, delta=0.2, do_photo=True):
    Delta /= dbu
    delta /= dbu  # new naming convention?
    for dp_lay in ['m2_nw_photo', 'm2_nw_ebeam']:
        cell.clear(lys[dp_lay])
    nw_region = as_region(cell, 'm2_nw')
    nw_compressed = fast_smoothed(nw_region)
    ebeam_region = fast_sized(nw_compressed, Delta + delta) - nw_region
    cell.shapes(lys.m2_nw_ebeam).insert(ebeam_region)
    if do_photo:
        phoas_region = as_region(cell, 'FLOORPLAN') - fast_sized(nw_compressed, Delta - delta)
        cell.shapes(lys.m2_nw_photo).insert(phoas_region)


@dpStep
def waveguide_sleeve(cell, Delta_nw_si=2.0, Delta=2.0, delta=0.2, do_photo=True):
    ''' Does a bulk-sleeve for waveguide full, but first adds it under nanowires.
        Endcaps end where the explicit waveguide ends.
        Recognizes the wg_deep_photo layer as a photolith-only layer (lower resolution, faster EBeam write)
    '''
    Delta_nw_si /= dbu
    Delta /= dbu
    delta /= dbu
    for dp_lay in ['wg_full_photo', 'wg_full_ebeam']:
        cell.clear(lys[dp_lay])

    # add silicon under the nanowires
    nw_compressed = fast_smoothed(as_region(cell, 'm2_nw'))
    wg_explicit = as_region(cell, 'wg_deep')
    nw_except_on_wg = nw_compressed - fast_sized(wg_explicit, Delta_nw_si)
    wg_all = fast_sized(nw_except_on_wg, Delta_nw_si) + wg_explicit

    # do the bulk-sleeve
    # wg_compressed = filter_large_polygons(wg_all)
    wg_compressed = fast_smoothed(wg_all)
    ebeam_region = fast_sized(wg_compressed, Delta + delta) - wg_all
    cell.shapes(lys.wg_full_ebeam).insert(ebeam_region)
    if do_photo:
        phoas_region = as_region(cell, 'FLOORPLAN') - fast_sized(wg_compressed, Delta - delta)
        try:
            phoas_region -= as_region(cell, 'wg_deep_photo')
        except KeyError: pass
        cell.shapes(lys.wg_full_photo).insert(phoas_region)


@dpStep
def ground_plane(cell, Delta_gp=15.0, points_per_circle=100, air_open=None):
    Delta_gp /= dbu
    cell.clear(lys.gp_photo)
    # Accumulate everything that we don't want to cover in metal
    gp_exclusion_things = pya.Region()
    for layname in ['wg_deep', 'wg_deep_photo', 'wg_shallow', 'm1_nwpad',
                    'm4_ledpad', 'm3_res', 'm5_wiring', 'm2_nw',
                    'GP_KO']:
        try:
            gp_exclusion_things += fast_smoothed(as_region(cell, layname))
        except KeyError: pass
    # Where ground plane is explicitly connected to wires, cut it out of the exclusion region
    gnd_explicit = as_region(cell, 'm5_gnd')
    gp_exclusion_tight = gp_exclusion_things - fast_sized(gnd_explicit, Delta_gp)
    # Inflate the buffer around excluded things and pour
    gp_exclusion_zone = fast_sized(gp_exclusion_tight, Delta_gp)
    gp_region = as_region(cell, 'FLOORPLAN') - gp_exclusion_zone

    # Connect to ground pads
    gp_region.merge()
    gp_region = fast_sized(gp_region, 1 / dbu)  # kill narrow spaces
    gp_region = fast_sized(gp_region, -2 / dbu)  # kill narrow widths
    gp_region = fast_sized(gp_region, 1 / dbu)
    gp_region += as_region(cell, 'm5_gnd')
    gp_region.round_corners(Delta_gp / 5, Delta_gp / 3, points_per_circle)
    gp_region = gp_region.smoothed(.001)  # avoid some bug in pya
    gp_region.merge()
    cell.shapes(lys.gp_photo).insert(gp_region)

    # Open up to the air
    if air_open is not None:
        Delta_air = 5
        fp_safe = as_region(cell, 'FLOORPLAN')
        air_rects = fp_safe - fp_safe.sized(0, -air_open / dbu, 0)
        air_region = air_rects & gp_region
        air_region = fast_sized(air_region, -Delta_air / dbu)
        air_region = fast_sized(air_region, 4 / dbu)  # kill narrow spaces
        air_region = fast_sized(air_region, -8 / dbu)  # kill narrow widths
        air_region = fast_sized(air_region, 4 / dbu)
        air_region.round_corners(Delta_gp / 5, Delta_gp / 3, points_per_circle)
        cell.shapes(lys.gp_v5).insert(air_region)


@dpStep
def metal_pedestal(cell, pedestal_layer='wg_full_photo', offset=0, keepout=None):
    metal_region = pya.Region()
    for layname in ['m5_wiring', 'm5_gnd', 'gp_photo']:
        try:
            metal_region += as_region(cell, layname)
        except: pass
    valid_metal = metal_region - fast_sized(as_region(cell, lys.wg_deep), offset / dbu)
    pedestal_region = fast_sized(valid_metal, offset / dbu)
    if keepout is not None:
        if not isinstance(keepout, (list, tuple)):
            keepout = [keepout]
        for ko_layer in keepout:
            pedestal_region -= as_region(cell, ko_layer)
    cell.shapes(lys[pedestal_layer]).insert(pedestal_region)


has_precomped = dict()
@dpStep
def precomp(cell, **kwargs):
    '''
        Arguments are keyed by layer name with the value of bias in microns, so for example

            precomp(TOP, wg_deep=0.05, nw_pad=-0.6)

        The behavior is to overwrite the layer.
        Problem if it runs twice is that it expands too much, so it stores which layers have been biased.
        Currently, it doesn't act on this, assuming that if you have run twice that you hit Undo in between.
    '''
    global has_precomped
    for layer_name, bias_um in kwargs.items():
        # do a check for repeated precomp
        if cell in has_precomped.keys():
            if layer_name in has_precomped[cell]:
                pya.MessageBox.info('Dataprep precomp', 'Warning: precompensating {} twice. '.format(layer_name) + '(in this process)\n'
                                    'If the last precomp was not undone with Ctrl-Z, this will turn out wrong', pya.MessageBox.Ok)
                pass#raise RuntimeError('At this time, precomp cannot be run twice on the same layer.')
            else:
                has_precomped[cell].add(layer_name)
        else:
            has_precomped[cell] = set([layer_name])

        # size the layer in place
        bias = bias_um / dbu
        layer_region = as_region(cell, layer_name)
        layer_region.size(bias)
        cell.clear(lys[layer_name])
        cell.shapes(lys[layer_name]).insert(layer_region)


@dpStep
def mask_map(cell, **kwargs):
    ''' lyp_file is relative to the yml file. If it is None, the same layer properties will be used.
        kwarg keys are destination layers and values are source layers, which can be a list

        There is a problem if you have 101 defined in your file and then another layer that is not defined.
    '''
    assert_valid_mask_map(kwargs)
    # If we need to make new layers,
    new_mask_index = 0
    available_mask_layers = list(range(100, 200))
    for occupied_layer in lys.values():
        if occupied_layer.layer in available_mask_layers:
            available_mask_layers.remove(occupied_layer.layer)
    # merging and moving to new layers
    for dest_layer, src_layers in kwargs.items():
        if not dest_layer in lys.keys():
            new_layinfo = pya.LayerInfo(available_mask_layers[new_mask_index], 0, dest_layer)
            lys[dest_layer] = new_layinfo
            cell.layout().layer(new_layinfo)

        if not isinstance(src_layers, list):
            src_layers = [src_layers]
        for src in src_layers:
            cell.copy(lys[src], lys[dest_layer])
        new_mask_index += 1


def assert_valid_mask_map(mapping):
    for dest_layer, src_layers in mapping.items():
        try:
            lys[dest_layer]
        except KeyError as err:
            message_loud('Warning: Destination layer [{}] not found in mask layerset. We will make it...'.format(dest_layer))
            pass  # This is allowed

        if not isinstance(src_layers, list):
            src_layers = [src_layers]
        for src in src_layers:
            try:
                lys[src]
            except KeyError as err:
                message_loud('Error: Source layer [{}] not found in existing designer or dataprep layerset.'.format(src))
                raise


@dpStep
def invert_tone(cell, layer):
    inverted = as_region(cell, 'FLOORPLAN') - as_region(cell, layer)
    cell.clear(lys[layer])
    cell.shapes(lys[layer]).insert(inverted)


@dpStep
def smooth_floating(cell, deviation=0.005):
    ''' Removes teeny tiny edges that sometimes show up in curved edges with angles 0 or 90 plus tiny epsilon
    '''
    for layer_name in lys.keys():
        layer_region = as_region(cell, layer_name)
        layer_region = fast_smoothed(layer_region, deviation)
        cell.clear(lys[layer_name])
        cell.shapes(lys[layer_name]).insert(layer_region)


@dpStep
def clear_nonmask(cell):
    ''' Gets rid of everything except 101--199. That is what we have decided are mask layers.
        Same as clear_others in mask_map
    '''
    for any_layer in lys.keys():
        lay = lys[any_layer]
        is_mask = (100 <= lay and lay < 200) or any_layer == 'FLOORPLAN'
        if not is_mask:
            cell.clear(lys[any_layer])


@dpStep
def align_corners(cell):
    ''' Puts little boxes in the corners so lithography tools all see the same
        Goes through all the layers present in the layout (does not depend on currently loaded layer properties)
    '''
    ly = cell.layout()
    all_layers = ly.layer_infos()

    # get the actual floorplan this time
    if cell.shapes(lys.FLOORPLAN).size() != 1:
        raise RuntimeError('align_corners needs a FLOORPLAN to work consisting of exactly one polygon')
    for fp in cell.shapes(lys.FLOORPLAN).each():
        fp_box = fp.dbbox()

    corner_mark = pya.DBox(0, 0, 1, 1)
    for marked_layer in all_layers:
        if marked_layer.name in ['FLOORPLAN']:  # put exceptions here
            continue
        if cell.shapes(ly.layer(marked_layer)).is_empty():
            continue
        # do some boolean here to shave off overhangs
        # layer_region = as_region(cell, marked_layer)
        # layer_region = layer_region & as_region(cell, 'FLOORPLAN')
        # cell.clear(lys[marked_layer])
        # cell.shapes(lys[marked_layer]).insert(layer_region)

        # put in the markers
        for north_south in (fp_box.top - 1, fp_box.bottom):
            for east_west in (fp_box.right - 1, fp_box.left):
                mark = corner_mark.moved(east_west, north_south)
                if marked_layer.name == 'DRC_exclude' or marked_layer.layer == 91:
                    mark = mark.enlarged(1, 1)
                cell.shapes(ly.layer(marked_layer)).insert(mark)


def assert_valid_dataprep_steps(step_list):
    ''' This runs before starting calculations to make sure there aren't typos
        that only show up after waiting for for all of the long steps
    '''
    if any(func_info_to_func_and_kwargs(func_info)[0] == 'add_library' for func_info in step_list):
        return

    # check function names
    for func_info in step_list:
        try:
            func = all_dpfunc_dict[func_info_to_func_and_kwargs(func_info)[0]]
        except KeyError as err:
            message_loud('Function "{}" not supported. Available are {}'.format(func_info[0], all_dpfunc_dict.keys()))
            raise

        # check mask layers
        if func is mask_map:
            assert_valid_mask_map(func_info_to_func_and_kwargs(func_info)[1])



'''
    Common functions used by the dataprep steps
'''
from __future__ import division, print_function, absolute_import
from functools import wraps
from lymask.utilities import active_technology, lys
from lygadgets import pya, message, message_loud

try:
    dbu = active_technology().dbu
except AttributeError:
    dbu = .001


# Metrics enum was added in v0.27
try:
    Euclidian = pya.Region.Euclidian
    Projection = pya.Region.Projection
except AttributeError:
    Euclidian = pya.Region.Metrics.Euclidian
    Projection = pya.Region.Metrics.Projection


def as_region(cell, layname):
    ''' Mostly a convenience brevity function.
        If a layer isn't in the layer set, return an empty region instead of crashing
        If a list, will return the union of the listed layers
    '''
    if isinstance(layname, (list, tuple)):
        union = pya.Region()
        for one_lay in layname:
            union += as_region(cell, one_lay)
        return union
    else:
        try:
            pya_layer = lys[layname]
        except KeyError:
            message(f'{layname} not found in layerset.')
            return pya.Region()
        return pya.Region(cell.shapes(pya_layer))


_thread_count = None
_tiles = None
def set_threads(thread_count, tiles=2):
    ''' Set to None to disable parallel processing '''
    global _thread_count, _tiles
    if thread_count == 1:
        thread_count = None
    _thread_count = thread_count
    _tiles = tiles


def _normal_smoothed(unfiltered_region, deviation=0.1):
    smoothed_region = unfiltered_region.dup()
    smoothed_region.merged_semantics = False
    smoothed_region.smooth(deviation / dbu)
    return smoothed_region


def fast_smoothed(unfiltered_region, deviation=0.1):
    ''' Removes any points that would change the shape by less than this deviation.
        This is used to significantly decrease the number of points prior to sizing

        Note: multicore does not work well with this, but it turns out to be pretty fast
    '''
    # if something goes wrong, you can fall back to regular here by uncommenting
    return _normal_smoothed(unfiltered_region, deviation)

    temp_region = unfiltered_region.dup()
    temp_region.merged_semantics = False

    output_region = pya.Region()
    tp = pya.TilingProcessor()
    tp.input('in1', temp_region)
    tp.output('out1', output_region)
    tp.queue("_output(out1, in1.smoothed({}))".format(deviation / dbu))
    # tp.tile_size(2000., 2000.)  # Not sure why this was here
    tp.tiles(_tiles, _tiles)
    tp.tile_border(5 * deviation, 5 * deviation)
    tp.threads = _thread_count
    tp.execute('Smoothing job')
    return output_region


def fast_sized(input_region, xsize):
    # if something goes wrong, you can fall back to regular here by uncommenting
    if _thread_count is None:
        return input_region.sized(xsize)
    else:
        output_region = pya.Region()
        tp = pya.TilingProcessor()
        tp.input('in1', input_region)
        tp.output('out1', output_region)
        tp.queue("_output(out1, in1.sized({}))".format(xsize))
        tp.tiles(_tiles, _tiles)
        tp.tile_border(2 * xsize, 2 * xsize)
        tp.threads = _thread_count
        tp.execute('Sizing job')
        return output_region


def fast_width(input_region, width, angle=90, min_projection=0):
    # if something goes wrong, you can fall back to regular here by uncommenting
    if _thread_count is None:
        return input_region.width_check(width, False, Projection, angle)
    else:
        output_edge_pairs = pya.EdgePairs()
        tp = pya.TilingProcessor()
        tp.input('in1', input_region)
        tp.output('out1', output_edge_pairs)
        if min_projection is None or min_projection == 0:
            min_projection = 'nil'
        tp.queue("_output(out1, in1.width_check({}, false, Region.Projection, {}, {}, nil))".format(width, angle, min_projection))

        border = 1.1 * width
        tp.tile_border(border, border)
        tp.tiles(_tiles, _tiles)
        tp.threads = _thread_count
        tp.execute('Width check job')
        return output_edge_pairs


def fast_space(input_region, spacing, angle=90, min_projection=0):
    # if something goes wrong, you can fall back to regular here by uncommenting
    if _thread_count is None:
        return input_region.space_check(spacing, False, Projection, angle, min_projection)
    else:
        output_edge_pairs = pya.EdgePairs()
        tp = pya.TilingProcessor()
        tp.input('in1', input_region)
        tp.output('out1', output_edge_pairs)
        # tp.queue("_output(out1, in1.space_check({}))".format(spacing))
        if min_projection is None or min_projection == 0:
            min_projection = 'nil'
        tp.queue("_output(out1, in1.space_check({}, false, Region.Projection, {}, {}, nil))".format(spacing, angle, min_projection))

        border = 1.1 * spacing
        tp.tile_border(border, border)
        tp.tiles(_tiles, _tiles)
        tp.threads = _thread_count
        tp.execute('Spacing check job')
        return output_edge_pairs


def fast_separation(r1, r2, exclude):
    # if something goes wrong, you can fall back to regular here by uncommenting
    if _thread_count is None:
        return r1.separation_check(r2, exclude)
    else:
        output_edge_pairs = pya.EdgePairs()
        tp = pya.TilingProcessor()
        tp.input('in1', r1)
        tp.input('in2', r2)
        tp.output('out1', output_edge_pairs)
        tp.queue("_output(out1, in1.separation_check(in2, {}))".format(exclude))

        border = 2 * exclude
        tp.tile_border(border, border)
        tp.tiles(_tiles, _tiles)
        tp.threads = _thread_count
        tp.execute('Separation check job')
        return output_edge_pairs


def turbo(input_region, meth_name, meth_args, tile_border=1, job_name='Tiling job'):
    ''' Speeds things up by tiling. Parameters are determined by _thread_count and _tiles
        if _thread_count is 1, it does not invoke tile processor at all
        tile_border is in microns. Recommended that you make it 1.1 * the critical dimension.
        args is a list.
    '''
    if not isinstance(meth_args, (list, tuple)):
        meth_args = [meth_args]
    if _thread_count is None:
        return getattr(input_region, meth_name)(*meth_args)
    else:
        output_region = pya.Region()
        tp = pya.TilingProcessor()
        tp.input('in1', input_region)
        tp.output('out1', output_region)

        clean_args = list()
        for arg in meth_args:
            if arg is True:
                clean_args.append('true')
            elif arg is False:
                clean_args.append('false')
            elif arg is None:
                clean_args.append('nil')
            elif isinstance(arg, (pya.Region, pya.EdgePairs)):
                tp.input('in2', arg)
                clean_args.append('in2')
            else:
                clean_args.append(str(arg))
        job_str = '_output(out1, in1.{}({}))'.format(meth_name, ', '.join(clean_args))
        tp.queue(job_str)

        tp.tile_border(tile_border, tile_border)
        tp.tiles(_tiles, _tiles)
        tp.threads = _thread_count
        tp.execute(job_name)
        return output_region


def rdb_create(rdb, cell, category, violations):
    rdb_cell = rdb.cell_by_qname(cell.name)
    trans_to_um = pya.CplxTrans(dbu)
    drc_exclude = as_region(cell, 'DRC_exclude')
    if isinstance(violations, pya.EdgePairs):
        # This handles edge pairs
        cleaned_violations = pya.EdgePairs()
        for ep in violations.each():
            edges = pya.Edges([ep.first, ep.second])
            # If drc_exclude touches it at all, don't add it
            if drc_exclude.interacting(edges).is_empty():
                cleaned_violations.insert(ep)
    else:
        # Everything else
        cleaned_violations = violations.outside(drc_exclude)
    rdb.create_items(rdb_cell.rdb_id(), category.rdb_id(), trans_to_um, cleaned_violations)




''' Command line entry points for dataprep and drc
'''
import argparse
from lymask import __version__
import textwrap
from lymask.invocation import batch_main, batch_drc_main

top_parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=textwrap.dedent('''\
        Available commands are
          dataprep: do some mask dataprep
          drc: do some design rule checking

        Type "lytest <command> -h" for help on specific commands
        '''))
top_parser.add_argument('command', type=str, choices=['dataprep', 'drc'], metavar='<command>')
top_parser.add_argument('args', nargs=argparse.REMAINDER)
top_parser.add_argument('-v', '--version', action='version', version='%(prog)s v{}'.format(__version__))

def cm_main():
    args = top_parser.parse_args()
    if args.command == 'dataprep':
        cm_dataprep(args.args)
    if args.command == 'drc':
        cm_drc(args.args)


def add_common_args(sub_parser):
    sub_parser.add_argument('infile', type=argparse.FileType('rb'),
                        help='the input gds file')
    sub_parser.add_argument('ymlspec', nargs='?', default=None,
                        help='YML file that describes the steps and parameters. Can be relative to technology')
    sub_parser.add_argument('-o', '--outfile', nargs='?', default=None,
                        help='The output file. Dataprep default is to tack "_proc" onto the end. DRC default is to put .lyrdb on the end')
    sub_parser.add_argument('-t', '--technology', nargs='?', default=None,
                        help='The name of technology to use. Must be visible in installed technologies')


dataprep_parser = argparse.ArgumentParser(prog='lymask dataprep' ,description="Command line mask dataprep")
add_common_args(dataprep_parser)

def cm_dataprep(args):
    dataprep_args = dataprep_parser.parse_args(args)
    batch_main(dataprep_args.infile.name, ymlspec=dataprep_args.ymlspec, outfile=dataprep_args.outfile, technology=dataprep_args.technology)


drc_parser = argparse.ArgumentParser(prog='lymask drc' ,description="Command line design rule check")
add_common_args(drc_parser)

def cm_drc(args):
    drc_args = drc_parser.parse_args(args)
    batch_drc_main(drc_args.infile.name, ymlspec=drc_args.ymlspec, outfile=drc_args.outfile, technology=drc_args.technology)



from __future__ import division, print_function, absolute_import
import os
import yaml
from lygadgets import isGUI, pya, message, message_loud, lyp_to_layerlist, patch_environment
from lygadgets.technology import Technology, klayout_last_open_technology
from lygadgets.gui_objects import gui_view, gui_active_technology


class objview(dict):
    ''' Supports both key and dot access/assignment '''
    __slots__ = ()

    def __getattr__(self, key):
        return self.__getitem__(key)
    def __setattr__(self, key, val):
        return self.__setitem__(key, val)
    def copy(self):
        return objview(super().copy())
    def deepcopy(self):
        return objview(deepcopy(self))


#: This global variable to be deprecated
_active_technology = None
def active_technology():
    ''' Gets active technology from GUI if in GUI, otherwise gives the stored variable, otherwise gives default last open
    '''
    global _active_technology
    if isGUI():
        return gui_active_technology()
    else:
        if _active_technology is None:
            _active_technology = Technology.technology_by_name(klayout_last_open_technology())
        return _active_technology


def set_active_technology(tech_name):
    if not Technology.has_technology(tech_name):
        raise ValueError('Technology not found. Available are {}'.format(Technology.technology_names()))
    if isGUI() and tech_name != gui_active_technology().name:
        raise RuntimeError('Cannot set technology via lymask while in GUI')
    global _active_technology
    _active_technology = Technology.technology_by_name(tech_name)
    reload_lys(tech_name, clear=True)
# end deprecation


def resolve_pya_tech(pya_tech_or_name=None):
    if pya_tech_or_name is None:
        pya_tech = active_technology()
    elif isinstance(pya_tech_or_name, str):
        pya_tech = Technology.technology_by_name(pya_tech_or_name)
    elif isinstance(pya_tech_or_name, pya.Technology):
        pya_tech = pya_tech_or_name
    else:
        raise TypeError(f'Cannot recognize {pya_tech_or_name} as an installed KLayout technology')
    return pya_tech


def tech_layer_properties(pya_tech=None):
    ''' Returns the file containing the main layer properties
    '''
    pya_tech = resolve_pya_tech(pya_tech)
    return pya_tech.eff_path(pya_tech.eff_layer_properties_file())


def tech_dataprep_layer_properties(pya_tech=None):
    ''' Returns the file containing the dataprep layer properties
    '''
    pya_tech = resolve_pya_tech(pya_tech)
    dataprep_path = pya_tech.eff_path('dataprep')
    for root, dirnames, filenames in os.walk(dataprep_path, followlinks=True):
        for filename in filenames:
            if filename.endswith('.lyp'):
                return os.path.join(root, filename)
    else:
        return tech_layer_properties()


def func_info_to_func_and_kwargs(func_info):
    ''' There are several ways to specify commands in the YML file. This parses them into function name and arguments
        It can be a list where first element is a function and second is a dict of kwargs.
        It can be a dict where key is function and value is dict of kwargs.
    '''
    if isinstance(func_info, str):
        func_name = func_info
        kwargs = dict()
    elif isinstance(func_info, list):
        message('Deprecation warning: spefifying a step as a list is going to go. Use dicts.')
        message(func_info)
        func_name = func_info[0]
        if len(func_info) == 1:
            kwargs = dict()
        elif len(func_info) == 2:
            kwargs = func_info[1]
        else:
            raise TypeError('Function not specified correctly as a list (needs two elements): {}'.format(func_info))
    elif isinstance(func_info, dict):
        if len(func_info.keys()) != 1:
            raise TypeError('Function not specified correctly as a dictionary (needs one key): {}'.format(func_info))
        for k, v in func_info.items():
            func_name = k
            kwargs = v
    else:
        raise TypeError('Function not specified correctly. Need str, list, dict: {}'.format(func_info))
    return func_name, kwargs


# --- Reading lyp files

class LayerSet(dict):
    ''' getitem returns the logical layer (integer) that can be used in pya functions,
        but you have to set the active_layout first

        It is keyed by the layer name such as 'si_wg',
        but it will succeed if the __getitem__ argument is
            1. integer, in which case, we assume it describes the physical layer to be converted to the logical layer
            2. pya.LayerInfo, in which case, it is ready to be processed by pya

        It also offers object-like attributes that can be gotten, set, and deleted like lys.si_wg.
        If you have a layer like "si_n+", you must use the __getitem__ style

        A note on iterating: this does not work because it side-steps the getitem code::

            for layname, lay in lys.items():

        Instead do this:

            for layname in lys.keys():
                lay = lys[layname]
    '''
    active_layout = None

    def __call__(self, *args, **kwargs):
        return self.__getitem__(args[0])

    def __getattr__(self, attrname):
        if attrname in self.keys():
            return self.__getitem__(attrname)

    def __setattr__(self, attrname, val):
        if attrname in ['active_layout']:  # exceptions
            self.__dict__[attrname] = val
        else:
            self.__setitem__(attrname, val)

    def __delattr__(self, attrname):
        del self[attrname]

    def get_as_LayerInfo(self, item):
        try:
            val = dict.__getitem__(self, item)
        except KeyError as err:
            if type(item) is int:
                val = pya.LayerInfo(item, 0)
            elif type(item) is pya.LayerInfo:
                val = item
            else:
                raise
        return val

    def __getitem__(self, item):
        val = self.get_as_LayerInfo(item)
        try:
            answer = self.active_layout.layer(val)
        except AttributeError as err:
            if "no attribute 'layer'" in err.args[0]:
                err.args = ('You didn\'t set active_layout for this LayerSet, so you can not get items from it', )
            raise
        return answer

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError('Key must be string. Got {}.'.format(type(key)))
        if not isinstance(value, pya.LayerInfo):
            raise TypeError('Value must be pya.LayerInfo. Got {}.'.format(type(value)))
        if not value.is_named():
            value.name = key
        dict.__setitem__(self, key, value)

    @classmethod
    def fromFile(cls, filename):
        new_obj = cls()
        all_layers = lyp_to_layerlist(filename)
        for one_layer in all_layers:
            try:
                group_members = one_layer['group-members']
            except KeyError:  # it is a real layer
                short_name = name2shortName(one_layer['name'])
                new_obj[short_name] = source2pyaLayerInfo(one_layer['source'])
            else:  # it is a group
                if not isinstance(group_members, list):
                    group_members = [group_members]
                for memb in group_members:
                    short_name = name2shortName(memb['name'])
                    new_obj[short_name] = source2pyaLayerInfo(memb['source'])
        return new_obj

    def append(self, other, doubles_ok=False):
        ''' When doubles_ok is True and there is a collision, the other takes precedence '''
        for layname in other.keys():
            if layname in self.keys() and not doubles_ok:
                raise ValueError('Layer is doubly defined: {}'.format(layname))
            self[layname] = other.get_as_LayerInfo(layname)

    def appendFile(self, filename, doubles_ok=False):
        other = LayerSet.fromFile(filename)
        other.active_layout = self.active_layout
        self.append(other, doubles_ok=True)


def name2shortName(name_str):
    ''' Good to have this function separate because
        it may differ for different naming conventions.

        Reassign with::

            lymask.utilities.name2shortName = someOtherFunction
    '''
    if name_str is None:
        raise IOError('This layer has no name')
    components = name_str.split(' - ')
    if len(components) > 1:
        short_name = components[1]
    else:
        short_name = components[0]
    return short_name


def source2pyaLayerInfo(source_str):
    layer_str = source_str.split('@')[0]
    layer, datatype = layer_str.split('/')
    return pya.LayerInfo(int(layer), int(datatype))


lys = LayerSet()
def reload_lys(technology=None, clear=False, dataprep=False):
    ''' Updates lys from the lyp files. Also updates the layer display in GUI mode.
        If any of the layers are already there, it does nothing.
        If no lyp is found, does nothing
    '''
    if technology is None:
        technology = active_technology()
    elif isinstance(technology, str):
        technology = Technology.technology_by_name(technology)

    if clear: lys.clear()
    try:
        lyp_file = tech_layer_properties(technology) if not dataprep else tech_dataprep_layer_properties(technology)
        lys.appendFile(lyp_file, doubles_ok=True)
    except (FileNotFoundError, AttributeError):
        message_loud('No lyp file found. Likely that technology hasn\'t loaded yet, or you don\'t have the standalone klayout')

    if isGUI():
        lv = gui_view()
        orig_list_index = lv.current_layer_list
        was_transacting = lv.is_transacting()
        if was_transacting:
            lv.commit()
        lv.load_layer_props(lyp_file)
        if was_transacting:
            lv.transaction('Bump transaction')
        lv.current_layer_list = orig_list_index

# reload_lys()


# --- Reading drc (yml) files

def tech_drc_properties(pya_tech=None, drc_filename=None):
    ''' Returns the file containing the dataprep layer properties
    '''
    pya_tech = resolve_pya_tech(pya_tech)
    drc_path = pya_tech.eff_path('drc')
    all_yml_files = []
    for root, dirnames, filenames in os.walk(drc_path, followlinks=True):
        for filename in filenames:
            if filename.endswith('.yml'):
                all_yml_files.append(os.path.join(root, filename))
    if len(all_yml_files) == 0:
        raise FileNotFoundError('No files matching drc/*.yml were found')
    if drc_filename is not None:
        if not '.' in drc_filename:
            drc_filename += '.yml'
        full_specified_path = os.path.join(root, drc_filename)
        if full_specified_path not in all_yml_files:
            raise FileNotFoundError(f'{len(all_yml_files)} yml files were found, but not {drc_filename}')
        return full_specified_path
    elif len(all_yml_files) > 1:
        all_bases = [os.path.basename(file) for file in all_yml_files]
        raise ValueError(f'Ambiguous DRC spec. Multiple found, so need to specify one: \n' + '\n'.join(all_bases))
    else:
        return all_yml_files[0]


def get_design_rules(pya_tech=None, drc_filename=None):
    ''' Access these like x = dr.width.RIB '''
    full_drc_path = tech_drc_properties(pya_tech, drc_filename)
    with open(full_drc_path, 'r') as fx:
        top_dict = yaml.load(fx, Loader=yaml.FullLoader)
    result_dict = objview(width=objview(), space=objview(), inclusion=objview(), exclusion=objview())
    for rule_info in top_dict:
        rule_name, subrule = func_info_to_func_and_kwargs(rule_info)
        if rule_name == 'width':
            result_dict.width[subrule['layer']] = float(subrule['value'])
        elif rule_name == 'space':
            result_dict.space[subrule['layer']] = float(subrule['value'])
        elif rule_name == 'inclusion':
            inners = subrule['inner']
            if not isinstance(inners, list):
                inners = [inners]
            outers = subrule['outer']
            if not isinstance(outers, list):
                outers = [outers]
            for inn in inners:
                for out in outers:
                    result_dict.inclusion[inn + '_' + out] = float(subrule['include'])
        elif rule_name == 'exclusion':
            lay1s = subrule['lay1']
            if not isinstance(lay1s, list):
                lay1s = [lay1s]
            lay2s = subrule['lay2']
            if not isinstance(lay2s, list):
                lay2s = [lay2s]
            for la1 in lay1s:
                for la2 in lay2s:
                    result_dict.exclusion[la1 + '_' + la2] = float(subrule['exclude'])
                    result_dict.exclusion[la2 + '_' + la1] = float(subrule['exclude'])
        else:
            pass
    return result_dict



from __future__ import division, print_function, absolute_import
from functools import wraps
from lygadgets import pya, isGUI, message, message_loud
from lygadgets.gui_objects import gui_view

from lymask.utilities import lys, LayerSet
from lymask.library import dbu, as_region, fast_sized, fast_smoothed, set_threads, rdb_create, fast_width, fast_space, fast_separation, turbo, Euclidian


all_drcfunc_dict = {}
def drcStep(step_fun):
    ''' Each step must accept one argument that is cell, plus optionals, and not return

        steps are added to all_steps *in the order they are defined*
    '''
    all_drcfunc_dict[step_fun.__name__] = step_fun
    return step_fun


__warned_about_flattening = False
@drcStep
def flatten(cell, rdb):
    global __warned_about_flattening
    if isGUI() and not __warned_about_flattening:
        message_loud('Warning: The flattening step modifies the layout, so be careful about saving.')
        __warned_about_flattening = True
    cell.flatten(True)

@drcStep
def make_rdbcells(cell, rdb):
    rdb.topcell = cell.name
    rdb_cell = rdb.create_cell(cell.name)


@drcStep
def processor(cell, rdb, thread_count=1, tiles=2, remote_host=None):
    if remote_host is not None:
        message_loud('Automatic remote hosting is not yet supported')
    set_threads(thread_count, tiles=tiles)


@drcStep
def drcX(cell, rdb, on_input=[], on_output=[], none=None):
    ''' DRC exclude handling. It takes lists of layers. There are three kinds
            on_input: removes that layer in the DRC_exclude regions. This is fast but dangerous if you create a small hole in some polygon
                This modifies the layout as you see it, so don't save.
            on_output (default): does the full computations but does not output edges that fall within the DRC_exclude parts
            none: output everything regardless of DRC_exclude
    '''
    if none is not None:
        raise RuntimeError('none-type DRC exclude is not supported yet.')
    for layer in on_input:
        pre_exclude = as_region(cell, layer)
        post_exclude = pre_exclude - as_region(cell, 'DRC_exclude')
        cell.clear(lys[layer])
        cell.shapes(lys[layer]).insert(post_exclude)
    for layer in on_output:
        pass  # good job you picked the default


@drcStep
def width(cell, rdb, layer, value, angle=90, min_projection=0):
    rdb_category = rdb.create_category('{}_Width'.format(layer))
    rdb_category.description = '{} [{:1.3f} um] - Minimum feature width violation'.format(layer, value)
    # message_loud('lymask doing {}'.format(rdb_category.name()))

    # do it
    polys = as_region(cell, layer)
    violations = fast_width(polys, value / dbu, angle, min_projection / dbu)
    # violations = polys.width_check(value / dbu, False, Euclidian, angle, None, None)
    # violations = turbo(polys, 'width_check', [value / dbu, False, Euclidian, angle, None, None],
    #                    tile_border=1.1*value, job_name='{}_Width'.format(layer))
    rdb_create(rdb, cell, rdb_category, violations)


@drcStep
def space(cell, rdb, layer, value, angle=90, min_projection=0):
    rdb_category = rdb.create_category('{}_Space'.format(layer))
    rdb_category.description = '{} [{:1.3f} um] - Minimum feature spacing violation'.format(layer, value)
    # message_loud('lymask doing {}'.format(rdb_category.name()))

    # do it
    polys = as_region(cell, layer)
    violations = fast_space(polys, value / dbu, angle, min_projection / dbu)
    # violations = turbo(polys, 'space_check', [value / dbu, False, Euclidian, angle, None, None],
    #                    tile_border=1.1*value, job_name='{}_Space'.format(layer))
    rdb_create(rdb, cell, rdb_category, violations)


@drcStep
def inclusion(cell, rdb, inner, outer, include):
    rdb_category = rdb.create_category('{} in {}'.format(inner, outer))
    rdb_category.description = '{} in {} [{:1.3f} um] - Minimum inclusion violation'.format(inner, outer, include)

    # do it
    rin = as_region(cell, inner)
    rout = as_region(cell, outer)
    small_rout = fast_sized(rout, -include / dbu)
    # Note: this could be parallelized, but it is easier I think than sizing
    violations = rin - small_rout

    rdb_create(rdb, cell, rdb_category, violations)


@drcStep
def exclusion(cell, rdb, lay1, lay2, exclude):
    rdb_category = rdb.create_category('{} from {}'.format(lay1, lay2))
    rdb_category.description = '{} from {} [{:1.3f} um] - Minimum exclusion violation'.format(lay1, lay2, exclude)

    # do it
    r1 = as_region(cell, lay1)
    r2 = as_region(cell, lay2)
    # r1.separation_check(r2, exclude / dbu)
    too_close = fast_separation(r1, r2, exclude / dbu)
    # This could be parallelized
    overlaps = r1 & r2

    rdb_create(rdb, cell, rdb_category, overlaps)
    rdb_create(rdb, cell, rdb_category, too_close)


def assert_valid_drc_steps(step_list):
    ''' This runs before starting calculations to make sure there aren't typos
        that only show up after waiting for for all of the long steps
    '''
    # check function names
    for func_info in step_list:
        try:
            func = all_drcfunc_dict[func_info[0]]
        except KeyError as err:
            message_loud('Function "{}" not supported. Available are {}'.format(func_info[0], all_drcfunc_dict.keys()))
            raise



''' This stuff only runs in GUI mode '''
from lygadgets import pya, message_loud
import glob
import os

from lymask.invocation import gui_main, gui_drc_main
from lymask.utilities import reload_lys, active_technology


def registerMenuItems():
    menu = pya.Application.instance().main_window().menu()
    s0 = "lymask_menu"
    if not(menu.is_menu(s0)):
        menu.insert_menu('macros_menu', s0, 'lymask')

    s1 = "lymask_menu.dataprep_menu"
    if not menu.is_menu(s1):
        menu.insert_menu('lymask_menu.end', 'dataprep', 'Mask Dataprep')

    s1 = "lymask_menu.drc_menu"
    if not menu.is_menu(s1):
        menu.insert_menu('lymask_menu.end', 'drc', 'Design Rule Check')


global item_counter
item_counter = 0
def _gen_new_action(func):
    ''' There a strange bug where pya.Actions get managed to the same location in memory.
        Same with the _Signals that are created when on_triggered is set.

        Assigning them to global variables with different names seems to work.
        It also works if you step through in a debugger.
        It does NOT work if you use locals()

        This function will create functions and action triggers correctly
    '''
    global item_counter
    item_str = 'action_item%s' % item_counter
    func_str = 'action_function%s' % item_counter
    globals()[item_str] = pya.Action()
    globals()[func_str] = func
    globals()[item_str].on_triggered = globals()[func_str]
    item_counter += 1
    return globals()[item_str]


def _gen_dataprep_action(dataprep_file):
    def wrapped():
        gui_main(dataprep_file)
    return _gen_new_action(wrapped)


def _gen_drc_action(drc_file):
    def wrapped():
        gui_drc_main(drc_file)
    return _gen_new_action(wrapped)


def _gen_path_action(tostr, pathstr):
    def wrapped():
        message_loud((
            'Path to {}\n'.format(tostr) +
            'Copy this path and paste wherever.\n\n' +
            pathstr
        ))
    return _gen_new_action(wrapped)


def dataprep_yml_to_menu(dataprep_file, menu_path='lymask_menu.dataprep'):
    ''' Takes a .yml file and adds a menu item that runs that file in lymask's drc or dataprep.
    '''
    menu = pya.Application.instance().main_window().menu()
    subloop_name = os.path.splitext(os.path.basename(dataprep_file))[0]
    if menu_path.endswith('dataprep'):
        action = _gen_dataprep_action(dataprep_file)
    elif menu_path.endswith('drc'):
        action = _gen_drc_action(dataprep_file)
    action.title = 'Run {}.yml'.format(subloop_name)
    if subloop_name == 'default':
        # action.shortcut = 'Shift+Ctrl+P'
        menu.insert_separator(menu_path + '.begin', 'SEP')
        menu.insert_item(menu_path + '.begin', subloop_name, action)
    else:
        menu.insert_item(menu_path + '.end', subloop_name, action)


def path_opener_to_menu(menu_path='lymask_menu.dataprep'):
    menu = pya.Application.instance().main_window().menu()
    tech = active_technology()
    if menu_path.endswith('dataprep'):
        category = 'dataprep'
    elif menu_path.endswith('drc'):
        category = 'drc'
    else:
        raise ValueError('Must give this a .dataprep or .drc menu_path')
    path = tech.eff_path(category)
    action = _gen_path_action(tech.name + ' ' + category, path)
    subloop_name = 'Path to {}:{}'.format(tech.name, category)
    action.title = subloop_name
    menu.insert_separator(menu_path + '.end', 'SEP')
    menu.insert_item(menu_path + '.end', subloop_name, action)


def reload_lymask_menu(category='dataprep', tech_name=None):
    if tech_name is None:
        tech = active_technology()
    else:
        tech = pya.Technology.technology_by_name(tech_name)
    menu = pya.Application.instance().main_window().menu()

    if category == 'dataprep':
        ymlfile_dir = tech.eff_path('dataprep')
        menu_path = 'lymask_menu.dataprep'
    elif category == 'drc':
        ymlfile_dir = tech.eff_path('drc')
        menu_path = 'lymask_menu.drc'

    # clear old ones
    for item in menu.items(menu_path):
        menu.delete_item(item)
    # insert new ones
    for ymlfile in glob.iglob(ymlfile_dir + '/*.yml'):
        dataprep_yml_to_menu(ymlfile, menu_path=menu_path)
    # Path opener
    path_opener_to_menu(menu_path=menu_path)




'''
    Entry points from GUI and API
'''
from __future__ import division, print_function, absolute_import
import os
import yaml
from lygadgets import pya, message, message_loud, Technology

from lygadgets.gui_objects import gui_view, gui_active_layout, gui_window, gui_active_technology
from lymask.utilities import active_technology, set_active_technology, \
                             tech_layer_properties, \
                             lys, reload_lys, func_info_to_func_and_kwargs
from lymask.dataprep_steps import all_dpfunc_dict, assert_valid_dataprep_steps
from lymask.drc_steps import all_drcfunc_dict, assert_valid_drc_steps


def _main(layout, ymlfile, tech_obj=None):
    with open(ymlfile) as fx:
        step_list = yaml.load(fx, Loader=yaml.FullLoader)
    reload_lys(tech_obj, dataprep=True)
    assert_valid_dataprep_steps(step_list)
    for func_info in step_list:
        func_name, kwargs = func_info_to_func_and_kwargs(func_info)
        message('lymask doing {}: {}'.format(func_name, kwargs))
        func = all_dpfunc_dict[func_name]
        for TOP_ind in layout.each_top_cell():
            # call it
            try:
                func(layout.cell(TOP_ind), **kwargs)
            except Exception as err:
                message_loud(str(err))
                raise
    return layout


def _drc_main(layout, ymlfile, tech_obj=None):
    with open(ymlfile) as fx:
        step_list = yaml.load(fx, Loader=yaml.FullLoader)
    if func_info_to_func_and_kwargs(step_list[0])[0] != 'make_rdbcells':
        step_list.insert(0, 'make_rdbcells')
    reload_lys(tech_obj, dataprep=True)
    # assert_valid_drc_steps(step_list)

    rdb = pya.ReportDatabase('DRC: {}'.format(os.path.basename(ymlfile)))
    rdb.description = 'DRC: {}'.format(os.path.basename(ymlfile))

    for func_info in step_list:
        func_name, kwargs = func_info_to_func_and_kwargs(func_info)
        message('lymask doing {}: {}'.format(func_name, kwargs))
        func = all_drcfunc_dict[func_name]
        for TOP_ind in layout.each_top_cell():
            try:
                func(layout.cell(TOP_ind), rdb, **kwargs)
            except Exception as err:
                message_loud(str(err))
                raise
    return rdb


def gui_main(ymlfile=None):
    layout = gui_active_layout()
    lys.active_layout = layout
    tech_obj = gui_active_technology()

    gui_view().transaction('Mask Dataprep')
    try:
        processed = _main(layout, ymlfile=ymlfile, tech_obj=tech_obj)
    finally:
        gui_view().commit()


def gui_drc_main(ymlfile=None):
    layout = gui_active_layout()
    lys.active_layout = layout
    tech_obj = gui_active_technology()

    lv = gui_view()
    lv.transaction('lymask DRC')
    try:
        rdb = _drc_main(layout, ymlfile=ymlfile, tech_obj=tech_obj)
    finally:
        lv.commit()
    rdix = lv.add_rdb(rdb)
    lv.show_rdb(rdix, lv.active_cellview().index())
    # Bring the marker browser window to the front
    gui_window().menu().action('tools_menu.browse_markers').trigger()


def batch_main(infile, ymlspec=None, technology=None, outfile=None):
    # covers everything that is not GUI
    if outfile is None:
        outfile = infile[:-4] + '_proc.oas'
    # Load it
    layout = pya.Layout()
    layout.read(infile)
    lys.active_layout = layout
    ymlfile = resolve_ymlspec(ymlspec, technology, category='dataprep')  # this also sets the technology
    tech_obj = active_technology()
    # Process it
    processed = _main(layout, ymlfile=ymlfile, tech_obj=tech_obj)
    # Write it
    processed.write(outfile)


def batch_drc_main(infile, ymlspec=None, technology=None, outfile=None):
    # covers everything that is not GUI
    if outfile is None:
        outfile = infile[:-4] + '.lyrdb'
    # Load it
    layout = pya.Layout()
    layout.read(infile)
    lys.active_layout = layout
    ymlfile = resolve_ymlspec(ymlspec, technology, category='drc')  # this also sets the technology
    tech_obj = active_technology()
    # Process it
    rdb = _drc_main(layout, ymlfile=ymlfile, tech_obj=tech_obj)
    # Write it
    rdb.save(outfile)
    # Brief report
    message('DRC violations:', rdb.num_items())
    message('Full report:', outfile)


def resolve_ymlspec(ymlspec=None, technology=None, category='dataprep'):
    ''' Find the yml file that describes the process. There are several options for inputs
        # Option 1: file path is specified directly
        # Option 2: search within a specified technology

        This also sets the active_technology stored in the module
    '''
    if ymlspec is not None and os.path.isfile(os.path.realpath(ymlspec)):
        # Option 1: file path is specified directly
        ymlfile = ymlspec
        if technology is not None:
            set_active_technology(technology)
        tech_obj = active_technology()
        if technology is None:
            message('Using the last used technology: {}'.format(tech_obj.name))
    else:
        # Option 2: search within a specified technology
        if technology is None:
            raise ValueError('When specifying a relative dataprep file, you must also provide a technology.')

        tech_obj = Technology.technology_by_name(technology)
        set_active_technology(technology)
        if ymlspec is None:
            # default dataprep test
            ymlspec = 'default'
        # find path to tech
        if not ymlspec.endswith('.yml'):
            ymlspec += '.yml'
        ymlfile = tech_obj.eff_path(os.path.join(category, ymlspec))
    if not os.path.isfile(ymlfile):
        raise FileNotFoundError('Could not resolve YAML specification: {}'.format(ymlspec))
    return ymlfile



import os, sys
import subprocess
import xmltodict
from collections import OrderedDict

import lymask
from lymask import batch_drc_main

from conftest import test_dir
drc_file = os.path.join(test_dir, 'tech', 'lymask_example_tech', 'drc', 'default.yml')
layout_file = os.path.join(test_dir, '2_drc_src.oas')
outfile = os.path.join(test_dir, '2_drc_run.lyrdb')
outfile_multithread = os.path.join(test_dir, '2_drc_multithread_run.lyrdb')
reffile = os.path.join(test_dir, '2_drc_answer.lyrdb')


class DRC_difference(Exception):
    pass


def assert_equal(rdb_file1, rdb_file2):
    ''' Errors if the rdbs are different.
        This is done with dictionaries not the XML text itself
        Order should not matter, so we use frozenset. In order to sort, we need to hash dicts,
        so we make frozendict. No need to mutate these, so it should be fine
    '''
    class frozendict(dict):
        def __hash__(self):
            return hash(tuple(sorted(self.items())))
        def __lt__(self, other):
            return hash(self) < hash(other)

    def strip_order(hierarchy):
        if isinstance(hierarchy, (dict, OrderedDict)):
            new_dict = {k: strip_order(v) for k,v in hierarchy.items()}
            return frozendict(new_dict)
        elif isinstance(hierarchy, (list, tuple)):
            new_list = [strip_order(el) for el in hierarchy]
            return frozenset(sorted(new_list))
        elif isinstance(hierarchy, str) and hierarchy.startswith('DRC:'):
            # Don't compare descriptions
            return 'DRC file'
        else:
            return hierarchy

    with open(rdb_file1, 'r') as fx:
        rdbspec1 = xmltodict.parse(fx.read(), process_namespaces=True)
    with open(rdb_file2, 'r') as fx:
        rdbspec2 = xmltodict.parse(fx.read(), process_namespaces=True)
    if strip_order(rdbspec1) != strip_order(rdbspec2):
        raise DRC_difference()


# This one need Technology working
def test_api():
    lymask.set_active_technology('lymask_example_tech')
    batch_drc_main(layout_file, ymlspec=drc_file, outfile=outfile)
    assert_equal(outfile, reffile)


def test_from_technology():
    batch_drc_main(layout_file, ymlspec='default', outfile=outfile, technology='lymask_example_tech')
    assert_equal(outfile, reffile)


def test_cm_from_tech():
    # this also checks that it defaults to default.yml
    command = ['lymask', 'drc']
    command += [layout_file]
    command += ['-o', outfile]
    command += ['-t', 'lymask_example_tech']
    subprocess.check_call(command)
    assert_equal(outfile, reffile)

def test_multithreaded():
    batch_drc_main(layout_file, ymlspec='multithreaded', outfile=outfile_multithread, technology='lymask_example_tech')
    assert_equal(outfile_multithread, reffile)

